@@ -15,6 +15,7 @@ class NumericEncoder:
         self._prepared = False
         self.is_target = is_target
         self.decode_log = False
+        self.extra_outputs = 0
 
     def prepare_encoder(self, priming_data):
         if self._prepared:
@@ -54,13 +55,14 @@ class NumericEncoder:
                 except:
                     real = None
             if self.is_target:
-                vector = [0] * 3
+                vector = [0] * (3 + 2*self.extra_outputs)
                 try:
-                    vector[0] = math.log(abs(real)) if abs(real) > 0 else - 20
-                    vector[1] = 1 if real < 0 else 0
+                    vector[0] = 1 if real < 0 else 0
+                    vector[1] = math.log(abs(real)) if abs(real) > 0 else - 20
                     vector[2] = real/self._mean
+                    # Note: This part here is just to have targets equall in size to the prediction size, these appended zeros won't be used anywhere
                 except Exception as e:
-                    vector = [0] * 3
+                    vector = [0] * (3 + 2*self.extra_outputs)
                     logging.error(f'Can\'t encode target value: {real}, exception: {e}')
 
             else:
@@ -99,10 +101,16 @@ class NumericEncoder:
                     real_value = pow(10,63)
                 else:
                     if decode_log:
-                        sign = -1 if vector[1] > 0.5 else 1
-                        real_value = math.exp(vector[0]) * sign
+                        sign = -1 if vector[0] > 0.5 else 1
+                        real_value = [math.exp(vector[i*2 + 1]) * sign for i in range(1 + self.extra_outputs)]
                     else:
-                        real_value = vector[2] * self._mean
+                        real_value = [vector[2*i + 2] * self._mean for i in range(1 + self.extra_outputs)]
+
+                    if self._type == 'int':
+                        real_value = [int(x) for x in real_value]
+                        
+                    if len(real_value) < 2:
+                        real_value = real_value[0]
             else:
                 if vector[0] < 0.5:
                     ret.append(None)
@@ -110,8 +118,8 @@ class NumericEncoder:
 
                 real_value = vector[3] * self._mean
 
-            if self._type == 'int':
-                real_value = round(real_value)
+                if self._type == 'int':
+                    real_value = round(real_value)
 
             ret.append(real_value)
         return ret
